/** @file cgenfunct.cpp convenience functions for cGenerator */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Copyright (C) 2007  Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "cgenfunct.h"

namespace faudes {

// controllability test for cGenerator plant
bool cIsControllable(
  const cGenerator& rPlantGen, 
  const vGenerator& rSupCandGen)
{
  return IsControllable(rPlantGen, rPlantGen.ControllableEvents(), rSupCandGen);
}


// supcon for cGenerators:
// uses and maintains controllablity from plant 
void cSupConNB(
  const cGenerator& rPlantGen, 
  const vGenerator& rSpecGen, 
  cGenerator& rResGen) {

  // prepare result
  cGenerator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.NewP();
  }

  // execute 
  SupConNB(rPlantGen, rPlantGen.ControllableEvents(),rSpecGen,*pResGen);

  // copy controllability of input alphabets
  pResGen->SetControllable(rPlantGen.ControllableEvents());

  // copy result
  if(pResGen != &rResGen) {
    rResGen = *pResGen;
    delete pResGen;
  }

}

// supcon for cGenerators
// uses and maintains controllablity from plant 
void cSupCon(
  const cGenerator& rPlantGen, 
  const vGenerator& rSpecGen, 
  cGenerator& rResGen) {

  // prepare result
  cGenerator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.NewP();
  }

  // execute
  SupCon(rPlantGen, rPlantGen.ControllableEvents(),rSpecGen,*pResGen);

  // copy controllability of input alphabets
  pResGen->SetControllable(rPlantGen.ControllableEvents());

  // copy result
  if(pResGen != &rResGen) {
    rResGen = *pResGen;
    delete pResGen;
  }

}

// parallel for cGenerators: sets the controllability 
// attributes of the events of the result.
void cParallel(
  const cGenerator& rGen1,
  const cGenerator& rGen2,
  cGenerator& rResGen) 
{

  // inputs have to agree on controllability of shared events:
#ifdef FAUDES_CHECKED
  EventSet inconsistent =
    (rGen1.ControllableEvents()*rGen2.UncontrollableEvents())
    +(rGen1.UncontrollableEvents()*rGen2.ControllableEvents());
    
    if(!inconsistent.Empty()) {
      std::stringstream errstr;
      errstr << "cParallel: inconsistent controllability flags";
      //throw Exception("TaGenerator::write", errstr.str(), 500);
    }
#endif

  // prepare result
  cGenerator* pResGen = &rResGen;
  if(&rResGen== &rGen1 || &rResGen== &rGen2) {
    pResGen= rResGen.NewP();
  }

  // make parallel composition of inputs
  Parallel(rGen1,rGen2,*pResGen);

  // copy controllability of input alphabets
  pResGen->SetControllable(rGen1.ControllableEvents());
  pResGen->SetControllable(rGen2.ControllableEvents());

  // copy result
  if(pResGen != &rResGen) {
    rResGen = *pResGen;
    delete pResGen;
  }
}

// cheap version of projection for cGenerators that sets the
// controllability attributes of the events of the result.
void cProject(
  const cGenerator& rGen,
  const EventSet& rProjectAlphabet,
  cGenerator& rResGen) 
{

  // prepare result
  cGenerator* pResGen = &rResGen;
  if(&rResGen== &rGen) {
    pResGen= rResGen.NewP();
  }

  // execute Project
  Project(rGen,rProjectAlphabet,*pResGen);

  // copy controllability of input alphabets
  pResGen->SetControllable(rGen.ControllableEvents()*rProjectAlphabet);

  // copy result
  if(pResGen != &rResGen) {
    rResGen = *pResGen;
    delete pResGen;
  }
}



}// namespace
